# BSD 2-Clause License

# Copyright 2022, Quoc-Tung Le

# Redistribution and use in source and binary forms, with or without modification, 
# are permitted provided that the following conditions are met:

# 1. Redistributions of source code must retain the above copyright notice, 
#     this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright notice, 
#     this list of conditions and the following disclaimer in the documentation 
#     and/or other materials provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, 
# OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY 
# OF SUCH DAMAGE.

import utils
import hierarchical_fact as fact
from scipy.linalg import hadamard

from iterative_method import training
import matplotlib.pyplot as plt
import matplotlib

import pandas as pd
import pickle
import time

import numpy as np

if __name__ == '__main__':
    learning_rate =[5e-1, 1e-1, 5e-2, 1e-2, 5e-3, 1e-3, 5e-4, 1e-4] 
    factors = range(3, 11)
    optimize = ['GD', 'GDMomentum', 'Adam']
    speed_dict = [dict(), dict()]

    df = [pd.DataFrame(), pd.DataFrame()]
    for method in optimize:
        speed_dict[0][method] = []
        speed_dict[1][method] = []
    speed_dict[0]['SVD'] = []
    speed_dict[1]['SVD'] = []
    
    for index in range(2):
        for num_factors in factors:
            if index == 0:
                support1 = utils.partial_prod_butterfly_supports(num_factors, 0, num_factors//2)
                support2 = utils.partial_prod_butterfly_supports(num_factors, num_factors//2, num_factors)
            else:
                support1 = utils.partial_prod_butterfly_supports(num_factors, 0, 1)
                support2 = utils.partial_prod_butterfly_supports(num_factors, 1, num_factors)
            size = 2 ** num_factors
            matrix = hadamard(size)
    
            begin = time.time()
            ft1, ft2 = fact.solve_DTO(support1, support2, matrix, type='real')
            end = time.time()

            speed_dict[index]['SVD'].append(end - begin)
        
            for method in optimize:
                best_time = np.inf
                best_error = np.inf
                for lr in learning_rate:
                    success, loss_array, time_array = training(matrix, support1, support2, lr = lr, option = method)
                    if not success:
                        continue
                    if (loss_array[-1] < 1e-10 or loss_array[-1] < best_error) and time_array[-1] < best_time: 
                        best_error = loss_array[-1]
                        best_time = time_array[-1]    
                speed_dict[index][method].append(best_time)
        
        df[index] = pd.DataFrame.from_dict(speed_dict[index])
        df[index].to_pickle("./data/type" + str(index) + ".pkl")
    
    speed_dict[0] = pd.read_pickle("./data/type0.pkl").to_dict('list')
    speed_dict[1] = pd.read_pickle("./data/type1.pkl").to_dict('list')
    
    fig, (ax1, ax2) = plt.subplots(1,2, figsize = (15, 5))
    for method in optimize:
        ax1.plot(2 ** np.array(factors), speed_dict[0][method], marker = '.', label=method)
    ax1.plot(2 ** np.array(factors), speed_dict[0]['SVD'], marker = '.', label = 'Algorithm 3.4', color = 'red')
    ax1.set_xscale("log")
    ax1.set_yscale("log")
    ax1.set_xlabel("Size")
    ax1.set_ylabel("Running time (s)")
    ax1.xaxis.set_major_formatter(matplotlib.ticker.ScalarFormatter())
    ax1.xaxis.set_minor_formatter(matplotlib.ticker.NullFormatter())
    ax1.set_xticks(2 ** np.array(factors))
    ax1.set_yticks([0.001, 0.01, 0.1, 1, 10, 100, 1000, 10000])
    ax1.legend()
    ax1.set_title("a)")
    ax1.grid()

    for method in optimize:
        ax2.plot(2 ** np.array(factors), speed_dict[1][method], marker = '.', label=method)
    ax2.plot(2 ** np.array(factors), speed_dict[1]['SVD'], marker = '.', label = 'Algorithm 3.4', color = 'red')
    ax2.set_xscale("log")
    ax2.set_yscale("log")
    ax2.set_xlabel("Size")
    ax2.set_ylabel("Running time (s)")
    ax2.xaxis.set_major_formatter(matplotlib.ticker.ScalarFormatter())
    ax2.xaxis.set_minor_formatter(matplotlib.ticker.NullFormatter())
    ax2.set_xticks(2 ** np.array(factors))
    ax2.set_yticks([0.001, 0.01, 0.1, 1, 10, 100, 1000, 10000])
    ax2.legend()
    ax2.set_title("b)")
    ax2.grid()
    plt.show()